
/*

	BuzzClock

	 1997 Steve Klingsporn
	moofie@pdicamillo.ne.highway1.com (temporary)

	    File:  TClockView.cpp

	Contains:  Clock view class implementation

	   Notes:  Version 1.1

*/

#ifndef _TCLOCKVIEW_H
#include "TClockView.h"
#endif

#ifndef _TCLOCKAPPLICATION_H
#include "TClockApplication.h"
#endif


/**
	String Constants
	
	Seems like we have linker problems if we put them in TClockView.h.
**/

const char 		*kApplicationSignatureString = "cclk";
const char 		*kSettingsFileName = "BuzzClock Settings";
const char		*kAboutItemName = "About BuzzClock";
const char		*kSecondsItemName = "Show Seconds";
const char		*k24HourItemName = "24-Hour Time";
const char		*kBeepItemName = "Beep on the Hour";
const char		*kCopyTimeStampItemName = "Copy as Time Stamp";
const char		*kQuitItemName = "Quit BuzzClock";


/**
	TClockView()
	
	Constructor.
**/

TClockView::TClockView(BRect frame)
		   :BView(frame, "", B_FOLLOW_LEFT | B_FOLLOW_TOP, 
			   	  B_WILL_DRAW | B_PULSE_NEEDED)
{
}


/**
	~TClockView()
	
	Destructor.  Writes the clock settings to the settings file
	and deletes allocated bitmaps.
**/

TClockView::~TClockView()
{
	WriteSettings();
	delete _closeBoxBitmaps[0];
	delete _closeBoxBitmaps[1];
	delete _clockBitmap;
}


/**
	AttachedToWindow()
	
	Initializes the default settings, sets view attributes, creates
	several bitmaps, and overrides the default settings with the settings
	in the settings file.
**/

void TClockView::AttachedToWindow()
{
	BRect	frame = Frame();
	
	_closeBoxState = false;
	_showSeconds = false;
	_militaryTime = false;
	_beepOnHour = true;
	_dateMode = false;
	_dateModeDelay = 0;

	SetViewColor(B_TRANSPARENT_32_BIT);

	//  Create the close box bitmaps
	
	_closeBoxBitmaps[0] = new BBitmap(BRect(0, 0, kCloseBoxSize, kCloseBoxSize), 
								      B_COLOR_8_BIT, FALSE);
	_closeBoxBitmaps[1] = new BBitmap(BRect(0, 0, kCloseBoxSize, kCloseBoxSize), 
									  B_COLOR_8_BIT, FALSE);
	
	_closeBoxBitmaps[0]->SetBits(kCloseBoxBitmapData, 144, 0, B_COLOR_8_BIT);
	_closeBoxBitmaps[1]->SetBits(kHilitedCloseBoxBitmapData, 144, 0, B_COLOR_8_BIT);
	
	//  Create the offscreen bitmap for the clock image
	
	_clockBitmap = new BBitmap(BRect(0, 0, frame.Width(), frame.Height()), B_COLOR_8_BIT, TRUE);
	_clockView = new BView(BRect(0, 0, frame.Width(), frame.Height()), "", B_FOLLOW_ALL, B_WILL_DRAW);
	_clockBitmap->AddChild(_clockView);
	
	//  Set the last time string
	
	TimeString(_lastTimeString);
	
	//  Create the main menu
	
	CreateMainMenu();
	
	//  Read in the application settings and update the bitmap
	
	ReadSettings();
	UpdateClockBitmap();
}


/**
	ReadSettings()
	
	Reads in the settings from the settings file.
**/

void TClockView::ReadSettings()
{
	BFile		*settingsFile;

	settingsFile = GetSettingsFile();
		
	if (settingsFile != NULL)
	{
		settingsFile->Open(B_READ_ONLY);
		
		settingsFile->Read((bool *)&_showSeconds, sizeof(bool));
		settingsFile->Read((bool *)&_militaryTime, sizeof(bool));
		settingsFile->Read((bool *)&_beepOnHour, sizeof(bool));
		
		settingsFile->Close();
		delete(settingsFile);
	}
}


/**
	WriteSettings()
	
	Writes the settings to the settings file.
**/

void TClockView::WriteSettings()
{
	BFile		*settingsFile;
	
	settingsFile = GetSettingsFile();
		
	if (settingsFile != NULL)
	{
		settingsFile->Open(B_READ_WRITE);
		
		settingsFile->Write((bool *)&_showSeconds, sizeof(bool));
		settingsFile->Write((bool *)&_militaryTime, sizeof(bool));
		settingsFile->Write((bool *)&_beepOnHour, sizeof(bool));
		
		settingsFile->Close();
		delete(settingsFile);
	}
}


/**
	GetSettingsFile()
	
	Returns the settings file, creating it if necessary.
**/

BFile *TClockView::GetSettingsFile()
{
	BVolume		bootVolume;
	BDirectory	*rootDirectory, *targetDirectory;
	BFile		*settingsFile = NULL;
	
	//  Navigate to the /system/settings directory.
	
	bootVolume = boot_volume();
	rootDirectory = new BDirectory();
	targetDirectory = new BDirectory();
	settingsFile = new BFile();
	
	bootVolume.GetRootDirectory(rootDirectory);
	rootDirectory->GetDirectory("system", targetDirectory);
	targetDirectory->GetDirectory("settings", targetDirectory);
	
	//  Attempt to locate the settings file.  If it doesn't exist,
	//  create it.
	
	if (targetDirectory->GetFile(kSettingsFileName, settingsFile) < B_NO_ERROR)
	{
		targetDirectory->Create(kSettingsFileName, settingsFile);
		WriteSettings();
	}
	
	return settingsFile;
}


/**
	CreateMainMenu()
	
	Creates the main menu.  This menu is used as the application's main
	menu, and is displayed when the user right-clicks in the view.
**/

void TClockView::CreateMainMenu()
{
	BMenuItem	*item;
	
	_mainMenu = new BPopUpMenu(kApplicationSignatureString, false, false, B_ITEMS_IN_COLUMN);
	
	item = new BMenuItem(kAboutItemName, new BMessage(B_ABOUT_REQUESTED), 0, 0);
	item->SetTarget(be_app);
	_mainMenu->AddItem(item);
	
	_mainMenu->AddItem(new BSeparatorItem());
	
	item = new BMenuItem(kCopyTimeStampItemName, new BMessage(MSG_COPY_TIME_STAMP), 0, 0);
	item->SetTarget(this);
	_mainMenu->AddItem(item);
	
	_mainMenu->AddItem(new BSeparatorItem());
	
	item = new BMenuItem(kSecondsItemName, new BMessage(MSG_TOGGLE_SECONDS), 0, 0);
	item->SetTarget(this);
	_mainMenu->AddItem(item);
	
	item = new BMenuItem(k24HourItemName, new BMessage(MSG_TOGGLE_24_TIME), 0, 0);
	item->SetTarget(this);
	_mainMenu->AddItem(item);
	
	item = new BMenuItem(kBeepItemName, new BMessage(MSG_TOGGLE_BEEP), 0, 0);
	item->SetTarget(this);
	_mainMenu->AddItem(item);
		
	_mainMenu->AddItem(new BSeparatorItem());
	
	item = new BMenuItem(kQuitItemName, new BMessage(B_QUIT_REQUESTED), 'Q', 0);
	item->SetTarget(be_app);
	_mainMenu->AddItem(item);
	
	//  In the rare case that the application menu actually
	//  becomes visible, we'll use this same main menu.
	
	be_app->SetMainMenu(_mainMenu);
}
	

/**
	AdjustMenuItems()
	
	Adjusts the mark for togglable menu items in the main menu.
**/

void TClockView::AdjustMenuItems()
{
	BMenuItem	*item;
	
	item = _mainMenu->FindItem(MSG_TOGGLE_SECONDS);
	if (_showSeconds)
		item->SetMarked(true);
	else
		item->SetMarked(false);
		
	item = _mainMenu->FindItem(MSG_TOGGLE_24_TIME);
	if (_militaryTime)
		item->SetMarked(true);
	else
		item->SetMarked(false);
		
	item = _mainMenu->FindItem(MSG_TOGGLE_BEEP);
	if (_beepOnHour)
		item->SetMarked(true);
	else
		item->SetMarked(false);
}


/**
	MouseDown()
	
	Called when the user clicks the mouse in the view.
**/

void TClockView::MouseDown(BPoint where)
{
	ulong		buttons;
	bool		previousCloseBoxState = _closeBoxState;
				
	GetMouse(&where, &buttons);
	
	if (buttons & B_SECONDARY_MOUSE_BUTTON && ! kCloseBoxFrame.Contains(where))
	{
		ConvertToScreen(&where);
		where.x -= kMainMenuOffset;
		AdjustMenuItems();
		_mainMenu->Go(where, true);
	}
	else
	{
		//  If the MouseDown point is in the close box frame
		//  rectangle, track the mouse until the button is released.
		//  Otherwise, toggle date mode.
		
		if (kCloseBoxFrame.Contains(where))
		{
			while (buttons)
			{				
				_closeBoxState = kCloseBoxFrame.Contains(where);
				
				if (_closeBoxState != previousCloseBoxState)
					Draw(kCloseBoxFrame);

				GetMouse(&where, &buttons);
				snooze(20000);
				previousCloseBoxState = _closeBoxState;
			}
		}
		else
		{
			_dateMode = ! _dateMode;
			_dateModeDelay = 0;
			UpdateClockBitmap();
			Draw(Frame());
		}
	}
		
	_closeBoxState = false;
	Invalidate(kCloseBoxFrame);
	
	if (kCloseBoxFrame.Contains(where)) 
	{
		be_app->PostMessage(B_QUIT_REQUESTED);
	}
}


/**
	TimeString()
	
	Puts the time in string format into timeString.  It returns
	a boolean that is true if we should beep on the hour.
**/

bool TClockView::TimeString(char *timeString)
{
	bool		evening;
	time_t		rawTime;
    tm			*timeInfo;
    char		suffix[8];
    bool		shouldBeep = false;
    
    time(&rawTime);
	evening = false;
	
	timeInfo = localtime(&rawTime);
	
	if (timeInfo->tm_hour > 11)
		evening = true;
		
	if (_showSeconds)
	{
		if (_militaryTime)
			sprintf(suffix, ":%02d", timeInfo->tm_sec);
		else
			sprintf(suffix, ":%02d %s", timeInfo->tm_sec, evening ? "PM" : "AM");
	}
	else
	{
		if (_militaryTime)
			sprintf(suffix, "");
		else
			sprintf(suffix, " %s", evening ? "PM" : "AM");
	}
	
	if (_beepOnHour)
	{
		if (timeInfo->tm_min == 0 && timeInfo->tm_sec == 0)
			shouldBeep = true;
	}
	
	sprintf(timeString, _militaryTime ? "%02d:%02d%s" : "%d:%02d%s", _militaryTime ? timeInfo->tm_hour : (timeInfo->tm_hour + 11) % 12 + 1, timeInfo->tm_min, suffix);
	
	return shouldBeep;
}


/**
	DateString()
	
	Puts the date in string format into dateString.
**/

void TClockView::DateString(char *dateString)
{
	time_t		rawTime;
    tm			*timeInfo;
	
	time(&rawTime);
	timeInfo = localtime(&rawTime);
	
	sprintf(dateString, "%d/%d/%02d", timeInfo->tm_mon + 1, timeInfo->tm_mday, timeInfo->tm_year);
}


/**
	UpdateClockBitmap()
	
	Updates the offscreen bitmap that we draw into the view.
**/

void TClockView::UpdateClockBitmap()
{
	char	timeString[32], dateString[10];	
	BRect	frame;

	frame = Frame();
	
	_clockBitmap->Lock();
	
	//  Draw the window title bar rectangle.
	
	_clockView->SetHighColor(kTitleBarBackgroundColor);
	_clockView->FillRect(BRect(frame.left, frame.top, 
							   frame.left + kTitleBarLeftMargin, frame.bottom));
	
	//  Draw the darker shadow lines around the title bar.
	
	_clockView->SetHighColor(kTitleBarShadowColor);
	_clockView->StrokeLine(BPoint(frame.left + kTitleBarLeftMargin, frame.top),
						   BPoint(frame.left + kTitleBarLeftMargin, frame.bottom));
	_clockView->StrokeLine(BPoint(frame.left, frame.bottom));
					
	//  Draw the lighter hilights around the title bar.
	
	_clockView->SetHighColor(kTitleBarHiliteColor);
	_clockView->StrokeLine(BPoint(frame.left + kTitleBarLeftMargin, frame.top),
						   BPoint(frame.left, frame.top));
	_clockView->StrokeLine(BPoint(frame.left, frame.bottom));
							   
	//  Draw the background panel.
	
	_clockView->SetHighColor(kPanelBackgroundColor);
	_clockView->FillRect(BRect(frame.left + kPanelLeftMargin, frame.top, 
							   frame.right, frame.bottom));
	
	//  Draw the darker shadow panel lines.
	
	_clockView->SetHighColor(kPanelShadowColor);
	_clockView->StrokeLine(BPoint(frame.right, frame.top),
						   BPoint(frame.right, frame.bottom));
	_clockView->StrokeLine(BPoint(frame.left + kPanelLeftMargin, frame.bottom));
							   
	//  Draw the panel hilights.
	
	_clockView->SetHighColor(kPanelHiliteColor);
	_clockView->StrokeLine(BPoint(frame.right, frame.top),
						   BPoint(frame.left + kPanelLeftMargin, frame.top));
	_clockView->StrokeLine(BPoint(frame.left + kPanelLeftMargin, frame.bottom));
		
	//  Set up the view for drawing text.
	
	_clockView->SetHighColor(kTextColor);
	_clockView->SetLowColor(kPanelBackgroundColor);
	_clockView->SetFontName("Emily");
	_clockView->MovePenTo(kTextPoint);
	
	//  If we're in date mode, draw the date.  Otherwise, draw the time.
	
	if (_dateMode)
	{
		DateString(dateString);
		_clockView->DrawString(dateString);
	}
	else
	{
		TimeString(timeString);
		_clockView->DrawString(timeString);
	}

	_clockView->Sync();
	_clockBitmap->Unlock();
}


/**
	Draw()
	
	Draws the offscreen bitmap into the view.
**/

void TClockView::Draw(BRect where)
{
#pragma unused(where)

	DrawBitmapAsync(_clockBitmap);
	
	//  Draw the close box.
	
	DrawBitmapAsync(_closeBoxBitmaps[_closeBoxState], BPoint(kCloseBoxFrame.left, kCloseBoxFrame.top));
}


/**
	Pulse()
	
	Called approximately every second.  If we're in date mode, we
	update a counter until we've paused for 3 seconds, then reset
	the view back to time mode.  If we're in time mode, we compare
	the previous time string to the current one, and update the
	bitmap and redraw the view if necessary.  If TimeString() returns
	true, we'll beep on the hour.
**/

void TClockView::Pulse()
{
	char	timeString[32];
	bool	shouldBeep;
	
	if (_dateMode)
	{
		_dateModeDelay++;
		
		//  Don't forget to beep on the hour if we're in
		//  date mode.
		
		if (_beepOnHour)
			if (TimeString(timeString))
				beep();
				
		if (_dateModeDelay >= kDateModeDelaySeconds)
		{
			_dateMode = false;
			_dateModeDelay = 0;
			UpdateClockBitmap();
			Draw(Frame());
		}
	}
	else
	{
		if (strcmp(timeString, _lastTimeString) != 0)
		{
			shouldBeep = TimeString(timeString);
			if (shouldBeep)
				beep();
				
			strcpy(_lastTimeString, timeString);
			UpdateClockBitmap();
			Draw(Frame());
		}
	}
}


/**
	MessageReceived()
	
	Called when we receive a Message.
**/

void TClockView::MessageReceived(BMessage *message)
{
	switch (message->what)
	{
		case MSG_TOGGLE_SECONDS:
			_showSeconds = ! _showSeconds;
			UpdateClockBitmap();
			Draw(Frame());
			break;
		
		case MSG_TOGGLE_24_TIME:
			_militaryTime = ! _militaryTime;
			UpdateClockBitmap();
			Draw(Frame());
			break;
			
		case MSG_TOGGLE_BEEP:
			_beepOnHour = ! _beepOnHour;
			break;
			
		case MSG_COPY_TIME_STAMP:
			CopyAsTimeStamp();
			break;
			
		default:
			BView::MessageReceived(message);
			break;
	}
}


/**
	CopyAsTimeStamp()
	
	Copies the current date and time (respecting seconds and time mode)
	to the clipboard.
**/

void TClockView::CopyAsTimeStamp()
{
	char	timeStampString[64], timeString[32], dateString[10];
	
	TimeString(timeString);
	DateString(dateString);
	
	sprintf(timeStampString, "%s, %s", dateString, timeString);
	
	be_clipboard->Lock();
	be_clipboard->Clear();
	be_clipboard->AddText(timeStampString);
	be_clipboard->Commit();
	be_clipboard->Unlock();
}
